# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import smtplib
import mimetypes
from typing import Tuple, List
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.audio import MIMEAudio
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart

from kudubot.connection.generic.Message import Message


class SmtpSender(object):
    """
    Class that handles the sending of emails via SMTP
    """

    credentials = ()
    """
    The credentials used to log on to the SMTP server
    """

    def __init__(self, credentials: Tuple[str, str, str, str, str]) -> None:
        """
        Constructor for the SmtpSender class that stores the credentials as a class variable

        :param credentials: The credentials to store
        :return: None
        """
        self.credentials = credentials

    def send_text_email(self, message: Message) -> None:
        """
        Sends an email message via SMTP
        :param message: The message to be send
        :return: None
        """
        body = MIMEText(message.message_body, 'plain')
        self.send_mime_part_email([body], message.message_title, message.address)

    def send_image_email(self, image_path: str, title: str, recipient: str) -> None:
        """
        Sends an embedded image via email and a title for it as the email subject

        :param image_path: the path to the image
        :param title: the image title
        :param recipient: the receiver of the email message
        :return: None
        """
        sub_type = self.guess_file_type(image_path)

        image_file = open(image_path, 'rb')  # Open the file in read-bytes mode
        image_data = image_file.read()
        image_file.close()

        image = MIMEImage(image_data, _subtype=sub_type)
        self.send_mime_part_email([image], title, recipient)

    def send_audio_email(self, audio_path: str, title: str, recipient: str) -> None:
        """
        Sends an audio file via email and a title for it as the email subject

        :param audio_path: the path to the audio file
        :param title: the audio title
        :param recipient: the receiver of the email message
        :return: None
        """
        sub_type = self.guess_file_type(audio_path)

        audio_file = open(audio_path, 'rb')  # Open the file in read-bytes mode
        audio_data = audio_file.read()
        audio_file.close()

        audio = MIMEAudio(audio_data, _subtype=sub_type)
        self.send_mime_part_email([audio], title, recipient)

    # noinspection PyTypeChecker
    def send_mime_part_email(self, mime_objects: List[MIMEBase], title: str, recipient: str) -> None:
        """
        Sends an email containing MIME Parts (like MIMEText or MIMEImage) and a title to a receiving
        email address

        :param mime_objects: a list of MIME objects to be sent
        :param title: the title of the email
        :param recipient: the receiver of the email
        :return: None
        """
        email_address, password, server, imap_port, smtp_port = self.credentials

        email_message = MIMEMultipart()
        email_message['From'] = email_address
        email_message['To'] = recipient
        email_message['Subject'] = title

        for part in mime_objects:
            email_message.attach(part)

        # Initialize SMTP Connection
        smtp = smtplib.SMTP_SSL("smtp." + server, int(smtp_port))
        smtp.set_debuglevel(0)
        smtp.login(email_address, password)

        # Send Email
        smtp.sendmail(email_address, recipient, email_message.as_string())
        smtp.quit()

    @staticmethod
    def guess_file_type(file_path: str) -> str:
        """
        Guesses the file tpye of a media file according to its file type extension

        :param file_path: the file to be checked
        :return: the media (sub)type
        """
        content_type, encoding = mimetypes.guess_type(file_path)
        if content_type is None or encoding is not None:
            content_type = 'application/octet-stream'
        return content_type.split('/', 1)[1]



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import configparser
from typing import List

from kudubot.servicehandlers.Service import Service
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class ServiceConfigParser(object):
    """
    Class that parses config files to determine which services to run.
    """

    # noinspection PyTypeChecker
    @staticmethod
    def read_config(all_services: List[Service], connection_identifier: str) -> List[Service]:
        """
        Reads the config file for the specific connection type and returns a list of plugins that
        are active according to the config file

        :param all_services: A list of all available services
        :param connection_identifier: A string that identifies the type of connection used
        :return: a list of the active plugins (according to the config file)
        """
        # Check config file
        config_file = os.path.join(LocalConfigChecker.config_directory, connection_identifier + "-services")

        # Sanity check
        config = open(config_file, 'r')
        config_contents = config.read()
        config.close()

        sane = True
        if not config_contents:
            sane = False
        if "[services]" not in config_contents:
            sane = False
        for service in all_services:
            if service.identifier not in config_contents:
                sane = False

        if not sane:
            ServiceConfigParser.write_standard_config(all_services, config_file)
            return list(all_services)
        else:
            config = configparser.ConfigParser()
            config.read(config_file)
            parsed_config = dict(config.items("services"))

            active_services = []

            try:
                for service in all_services:
                    if parsed_config[service.identifier] == "1":
                        active_services.append(service)
                return active_services
            except KeyError:
                ServiceConfigParser.write_standard_config(all_services, config_file)

    # noinspection PyTypeChecker
    @staticmethod
    def write_standard_config(all_services: List[Service], config_file: str) -> None:
        """
        Writes a service config file with all services enabled

        :param all_services: A list of all available services
        :param config_file: The config file to write to
        :return: None
        """
        config = open(config_file, "w")
        config.write("[services]\n")

        for service in all_services:
            config.write(service.identifier + " = 1\n")

        config.close()



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import requests
from bs4 import BeautifulSoup
from typing import List, Dict

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class KickTippService(Service):
    """
    The KickTippService Class that extends the generic Service class.
    The service parses www.kicktipp.de to get a kicktipp group's current standings
    """

    identifier = "kicktipp"
    """
    The identifier for this service
    """

    help_description = {"en": "/kicktipp\tFetches Kicktipp Community Tables\n"
                              "syntax: /kicktipp <kicktipp-community>",
                        "de": "/kicktipp\tZeigt Kicktipp Community Tabellen\n"
                              "syntax: /kicktipp <kicktipp-community>"}
    """
    Help description for this service.
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        kicktipp_community = message.message_body.split("/kicktipp ", 1)[1]
        reply = self.get_kicktipp_info(kicktipp_community)

        reply_message = self.generate_reply_message(message, "Kicktipp", reply)
        if self.connection.identifier in ["whatsapp", "telegram"]:
            self.send_text_as_image_message(reply_message)
        else:
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return re.search(r"^/kicktipp [a-z]+(\-|[a-z]+)*[a-z]+$", message.message_body.lower())

    def get_kicktipp_info(self, kicktipp_community: str) -> str:
        """
        Gets the current kicktipp table standings from www.kicktipp.de for a specified kicktipp
        community.

        :param kicktipp_community: the kicktipp community to be checked
        :return: the kicktipp table standings
        """
        # Get the data from kicktipp.de
        html = requests.get("http://www.kicktipp.de/" + kicktipp_community + "/tippuebersicht").text
        soup = BeautifulSoup(html, "html.parser")
        names = soup.select(".mg_class")
        normal_scores = soup.select(".pkt")
        winner_scores = soup.select(".pkts")

        users = []

        normal_score_counter = 0
        winner_score_counter = 0

        for user in range(0, len(names)):
            user_dictionary = {'position': str(user + 1),
                               'name': names[user].text}
            if re.search(r"[0-9]+,[0-9]+", normal_scores[normal_score_counter].text):
                # If the first .pkt element for the user is already a float, it means that the user is a matchday
                # winner. Since winner's matchday points are stored in .pkts elements, we need to get the matchday
                # points of the winner from the .pkts elements and the rest from the .pkt elements
                user_dictionary['day_points'] = winner_scores[winner_score_counter].text
                user_dictionary['wins'] = normal_scores[normal_score_counter].text
                user_dictionary['total_points'] = normal_scores[normal_score_counter + 1].text
                winner_score_counter += 1
                normal_score_counter += 2
            else:
                # Otherwise, if the user is not a winner, just use the .pkt elements for everything
                user_dictionary['day_points'] = normal_scores[normal_score_counter].text
                user_dictionary['wins'] = normal_scores[normal_score_counter + 1].text
                user_dictionary['total_points'] = normal_scores[normal_score_counter + 2].text
                normal_score_counter += 3
            users.append(user_dictionary)

        return self.format_table(users)

    def format_table(self, users: List[Dict[str, str]]) -> str:
        """
        Formats the table for the current connection type

        :param users: A list of dictionaries containing the information for all users.
        :return: the table, formatted
        """
        longest_name = 0
        for user in users:
            if len(user["name"]) > longest_name:
                longest_name = len(user["name"])

        if self.connection.identifier in ["whatsapp", "telegram"]:
            formatted_table = "Pos Name         Pts Wins Tot\n"

            for user in users:
                formatted_table += user["position"].ljust(4)
                formatted_table += user["name"].ljust(12)
                formatted_table += user["day_points"].ljust(4)
                formatted_table += user["wins"].ljust(6)
                formatted_table += user["total_points"] + "\n"

        else:
            formatted_table = "Pos " + "Name".ljust(longest_name + 1) + " Pts  Wins  Tot\n"

            for user in users:
                formatted_table += user["position"].ljust(4)
                formatted_table += user["name"].ljust(longest_name + 1)
                formatted_table += user["day_points"].ljust(5)
                formatted_table += user["wins"].ljust(7)
                formatted_table += user["total_points"] + "\n"

        return formatted_table



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""
# imports
import re
import tvdb_api
from tvdb_exceptions import tvdb_episodenotfound, tvdb_seasonnotfound, tvdb_shownotfound
from requests.exceptions import ConnectionError

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class TvdbService(Service):
    """
    The TvdbService Class that extends the generic Service class.
    The service gets tv show nformation from thetvdb.com
    """

    identifier = "tvdb"
    """
    The identifier for this service
    """

    help_description = {"en": "/tvdb\tSends episode name of an episode from TVDB\n"
                              "syntax: /tvdb <show> s<season> e<episode>",
                        "de": "/tvdb\tSchickt den Episodennamen einer Episode auf TVDB\n"
                              "syntax: /tvdb <show> s<staffel> e<episode>"}
    """
    Help description for this service.
    """

    episode_not_found_error = {"en": "Episode not found on the TV database (thetvdb.com)",
                               "de": "Episode nicht auf der TV Database gefunden (thetvdb.com)"}

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        tv_show = message.message_body.split(" ", 1)[1].rsplit(" s", 1)[0]
        season = int(message.message_body.rsplit("s", 1)[1].rsplit(" e", 1)[0])
        episode = int(message.message_body.rsplit("e", 1)[1])

        reply = self.get_tvdb_info(tv_show, season, episode)

        reply_message = self.generate_reply_message(message, "Tvdb", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return re.search(r"^/(tvdb) ([^ ]+| )+ s[0-9]{1,2} e[0-9]{1,4}$", message.message_body.lower())

    def get_tvdb_info(self, show: str, season: int, episode: int) -> str:
        """
        Gets information for the given tv show parameters

        :param show: the show to check
        :param season: the season to check
        :param episode: the episode to check
        :return: the received information from thetvdb.com
        """
        try:
            tvdb = tvdb_api.Tvdb()
            episode_info = tvdb[show][season][episode]
            return episode_info['episodename']
        except (tvdb_episodenotfound, tvdb_seasonnotfound, tvdb_shownotfound, ConnectionError):
            return TvdbService.episode_not_found_error[self.connection.last_used_language]



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from nose.tools import with_setup
from nose.tools import assert_false
from nose.tools import assert_true

from kudubot.connection.generic.Message import Message
from kudubot.services.internet_services.CinemaService import CinemaService


# noinspection PyMethodMayBeStatic
class TestCinemaService(object):
    """
    A Unit Test Class for a Service class
    """

    correct_messages = ["/cinema karlsruhe", "/cinema karlsruhe 10", "/cinema new york city 20", "/cinema cape town",
                        "/kino place"]
    incorrect_messages = ["/cinem karlsruhe", "/cinema karlsruhe -1", "balaalal/cinema karlsruhe shjsncksjsaax",
                          "balaalal/cinema karlsruhe---------"]
    service = CinemaService

    @classmethod
    def setup_class(cls):
        """
        Sets up the test class
        """
        pass

    @classmethod
    def teardown_class(cls):
        """
        Tears down the test class
        """
        pass

    def setup(self):
        """
        Sets up a test
        """
        pass

    def teardown(self):
        """
        Tears down a test
        """
        pass

    @with_setup(setup, teardown)
    def test_regex(self):
        """
        Tests the service's regex check
        """
        for message in self.correct_messages:
            message_object = Message(message_body=message, address="")
            print("Testing correct Regex for: " + message)
            assert_true(self.service.regex_check(message_object))
        for message in self.incorrect_messages:
            message_object = Message(message_body=message, address="")
            assert_false(self.service.regex_check(message_object))
            print("Testing incorrect Regex for: " + message)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class HelpService(Service):
    """
    The HelpService Class that extends the generic Service class.
    The service lists the help strings for all running dervices
    """

    identifier = "help"
    """
    The identifier for this service
    """

    protected = True
    """
    May not be disabled
    """

    help_description = {"en": "No Help Description Available",
                        "de": "Keine Hilfsbeschreibung verfügbar"}
    """
    Help description for this service. No description since this is the class generating the help\
    messages anyway.
    """

    help_keywords = {"help": "en",
                     "hilfe": "de"}
    """
    Keywords that trigger the help service, which can also be used to determine the language
    """

    instruction_message = {"en": "\nFor detailed instructions, enter \n/help <service-name> "
                                 "\nor \n/help <service-index>",
                           "de": "\nFür detailierte Anweisungen, geb \n/hilfe <service-name>"
                                 "\noder \n/hilfe <service-index> ein"}
    """
    The instruction message at the end of the help message
    """

    service_not_found_warning = {"en": ("Sorry, the service \"", "\" was not found"),
                                 "de": ("Sorry, der Service \"", "\" wurde nicht gefunden")}
    """
    Message shown when the help service didn't find a specified service
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        language = self.help_keywords[message.message_body.split("/")[1].split(" ")[0]]

        try:
            selected_service = message.message_body.split(" ")[1]
        except IndexError:
            selected_service = None

        indexed_services = {}
        for service_count in range(0, len(self.connection.service_manager.all_services)):
            indexed_services[service_count] = self.connection.service_manager.all_services[service_count]

        if selected_service is None:
            reply = "Services\n\n"

            for service_count in range(0, len(self.connection.service_manager.all_services)):
                reply += str(service_count + 1) + ": " + indexed_services[service_count].identifier  # + "\n"
                if indexed_services[service_count].protected:
                    reply += " " + "🔐"  # Lock emoji
                if indexed_services[service_count] in self.connection.service_manager.active_services:
                    reply += " " + "👍🏻"  # thumbs up emoji
                else:
                    reply += " " + "👎🏻"  # thumbs down emoji
                reply += "\n"

            reply += self.instruction_message[language]

        else:
            reply = ""
            try:
                reply = self.get_service_description(indexed_services[int(selected_service) - 1], language)
            except (ValueError, KeyError):
                for service in self.connection.service_manager.all_services:
                    if service.identifier == selected_service:
                        reply = self.get_service_description(service, language)
                if not reply:
                    reply = self.service_not_found_warning[language][0] + selected_service \
                            + self.service_not_found_warning[language][1]

        reply_message = self.generate_reply_message(message, "Help", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        match = False

        for keyword in HelpService.help_keywords:
            if message.message_body.startswith("/" + keyword + " ") or message.message_body == "/" + keyword:
                match = True

        return match

    @staticmethod
    def get_service_description(service: Service, language: str) -> str:
        """
        Method that returns the help description of a service in a particular language.
        If the selected language has no description, the string
        "No help available for this language"
        will be returned instead

        :param service: the service for which the help message should be returnes
        :param language: the language to be returned
        :return: the description in that language
        """
        try:
            return service.help_description[language.lower()]
        except KeyError:
            return "No help available for this language"



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
import time
import calendar
import datetime
from typing import Tuple, Dict, List

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class ReminderService(Service):
    """
    The ReminderService Class that extends the generic Service class.
    The service offers reminder at specific or relative times
    """

    identifier = "reminder"
    """
    The identifier for this service
    """

    help_description = {"en": "/remind\tSaves a reminder and sends it back at the specified time\n"
                              "syntax: /remind \"<message>\" <time>\n"
                              "time syntax: <YYYY-MM-DD-hh-mm-ss>\n"
                              "or: <amount> [years|months|days|hours|minutes|seconds]\n"
                              "or: tomorrow\n\n"
                              "/remind list\tLists all reminders currently stored\n"
                              "/remind delete <index>\tDeletes the reminder at the given index\n\n"
                              "All times are stored in UTC",
                        "de": "/remind\tSpeichert eine Erinnerung und verschickt diese zum angegebenen Zeitpunkt\n"
                              "syntax: /erinner \"<nachricht>\" <zeit>\n"
                              "zeit syntax: YYYY-MM-DD-hh-mm-ss\n"
                              "oder: <anzahl> [jahre|monate|tage|stunden|minuten|sekunden]\n"
                              "oder: morgen\n\n"
                              "/erinner list\tListet alle Erinnerungen\n"
                              "/erinner delete <index>\tDeletes the reminder at the given index\n\n"
                              "Alle Zeiten werden in UTC gespeichert"}
    """
    Help description for this service.
    """

    has_background_process = True
    """
    Indicates that the serice has a background process
    """

    message_stored_reply = {"en": "Message Stored",
                            "de": "Nachricht gespeichert"}
    """
    Reply to signal the user that the reminder was successfully stored
    """

    reminder_directory = "reminder"
    """
    The directory in which the reminders are stored
    """

    remind_keywords = {"remind": "en",
                       "erinner": "de"}
    """
    Keywords for the remind command
    """

    time_adder_keywords = {"years": ["en", "years"],
                           "jahre": ["de", "years"],
                           "months": ["en", "months"],
                           "monate": ["de", "months"],
                           "days": ["en", "days"],
                           "tage": ["de", "days"],
                           "hours": ["en", "hours"],
                           "stunden": ["de", "hours"],
                           "minutes": ["en", "minutes"],
                           "minuten": ["de", "minutes"],
                           "seconds": ["en", "seconds"],
                           "sekunden": ["de", "seconds"]}
    """
    Keywords for specific time sizes
    """

    tomorrow_keywords = {"tomorrow": "en",
                         "morgen": "de"}
    """
    Keywords that define the tomorrow option
    """

    remider_time = {"years": 0,
                    "months": 0,
                    "days": 0,
                    "hours": 0,
                    "minutes": 0,
                    "seconds": 0}
    """
    The time when the reminder shoul be activated
    """

    list_keywords = {"list": "en",
                     "liste": "de"}
    """
    Keywords for the list command
    """

    no_reminders_stored = {"en": "No reminders stored",
                           "de": "Keine Erinnerungen gespeichert"}
    """
    Message sent if the user requests a list of his reminders, but there are none
    """

    reminder_delete_out_of_bounds = {"en": "No reminder with that index stored",
                                     "de": "Keine Erinnerung mit diesem Index."}
    """
    Message sent when the delete command's index is out of bounds
    """

    delete_file_success = {"en": "Successfully deleted reminder",
                           "de": "Erinnerung erfolgreich gelöscht"}
    """
    Success message when deleting an reminder
    """

    delete_keywords = {"delete": "en",
                       "löschen": "de"}
    """
    Keywords for the delete command
    """

    def initialize(self) -> None:
        """
        Constructor extender for the Renamer class that initializes a directory for the reminder files

        :return: None
        """
        self.reminder_directory = \
            os.path.join(LocalConfigChecker.services_directory, self.connection.identifier, self.reminder_directory)
        LocalConfigChecker.validate_directory(self.reminder_directory)

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        user_input = message.message_body.lower()
        delete_key = None
        for key in self.delete_keywords:
            if key in user_input:
                delete_key = key
        list_key = None
        for key in self.list_keywords:
            if key in user_input:
                list_key = key

        if delete_key is not None:  # In other words: if "delete" in message.message_body.lower()
            language = self.delete_keywords[delete_key]
            index = int(message.message_body.split(delete_key + " ")[1])
            reply = self.delete_reminder_for_user(message.address, index, language)

        elif list_key is not None:
            language = self.list_keywords[list_key]
            reply = self.get_user_reminders_as_string_from(message.address, language)

        else:
            language, reminder_options, reminder_message = self.parse_user_input(message.message_body)
            reminder_time = self.determine_reminder_time(reminder_options)
            reminder_time = self.normalize_time(reminder_time)

            self.store_reminder(reminder_time, reminder_message, message.address)
            reply = self.message_stored_reply[language]

        reply_message = self.generate_reply_message(message, "Reminder", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/" + Service.regex_string_from_dictionary_keys([ReminderService.remind_keywords])
        regex += " \"[^\"]+\" ([0-9]+ " + \
                 Service.regex_string_from_dictionary_keys([ReminderService.time_adder_keywords])
        regex += "|" + Service.regex_string_from_dictionary_keys([ReminderService.tomorrow_keywords])
        regex += "|[0-9]{4}-[0-9]{2}-[0-9]{2}(-[0-9]{2}-[0-9]{2}-[0-9]{2})?)$"

        list_regex = "^/" + Service.regex_string_from_dictionary_keys([ReminderService.remind_keywords]) + " "
        list_regex += Service.regex_string_from_dictionary_keys([ReminderService.list_keywords]) + "$"

        delete_regex = "^/" + Service.regex_string_from_dictionary_keys([ReminderService.remind_keywords]) + " "
        delete_regex += Service.regex_string_from_dictionary_keys([ReminderService.delete_keywords]) + " [0-9]+$"

        return re.search(re.compile(regex), message.message_body.lower())\
            or re.search(re.compile(list_regex), message.message_body.lower())\
            or re.search(re.compile(delete_regex), message.message_body.lower())

    def parse_user_input(self, user_input: str) -> Tuple[str, str, str]:
        """
        Parses the user input and determines the used language, the reminder to be stored,
        as well as th options defining when the reminder will be activated

        :param user_input: the user input to be parsed
        :return: language, options, reminder text
        """

        remind_keyword, options = user_input.split(" ", 1)  # Split message into /remind and the rest of the message
        language = self.remind_keywords[remind_keyword.lower().split("/")[1]]  # Get the language using /remind keyword

        # Split message into three parts:
        # junk: empty string, stuff that comes before the first quotation mark
        # reminder_message: The message nested between quotation marks
        # options: the options after the reminder string
        junk, reminder_message, options = options.split("\"", 2)

        options = options.lower().lstrip()  # Remove leading whitespace

        return language, options, reminder_message

    def determine_reminder_time(self, options: str) -> Dict[str, int]:
        """
        Determines the reminder time from the given options

        :param options: the options to be parsed
        :return: the reminder time as a dictionary
        """
        reminder_time = self.get_time()

        if options in self.tomorrow_keywords:
            reminder_time["days"] += 1

        elif re.search(r"[0-9]{4}-[0-9]{2}-[0-9]{2}(-[0-9]{2}-[0-9]{2}-[0-9]{2})?", options):
            selected_time = options.split("-")
            reminder_time["years"] = int(selected_time[0])
            reminder_time["months"] = int(selected_time[1])
            reminder_time["days"] = int(selected_time[2])
            if len(selected_time) > 3:
                reminder_time["hours"] = int(selected_time[3])
                reminder_time["minutes"] = int(selected_time[4])
                reminder_time["seconds"] = int(selected_time[5])

        else:
            modifier, key = options.split(" ")
            reminder_time[key] += int(modifier)

        return reminder_time

    def store_reminder(self, reminder_time: Dict[str, int], reminder_message: str, reminder_address: str) -> None:
        """
        Stores a reminder in the reminder directory as a reminder file.

        :param reminder_time: the time the reminder activates
        :param reminder_message: the reminder text itself
        :param reminder_address: the address to where the reminder will be sent
        :return: None
        """
        timestamp = ReminderService.convert_time_to_utc_timestamp(reminder_time)

        # emulate a do-while loop
        # The loop makes sure that no existing file is overwritten
        counter = 0
        while True:
            reminder_filename = str(timestamp) + "#" + str(counter) + "@" + reminder_address
            reminder_file = os.path.join(self.reminder_directory, reminder_filename)
            if not os.path.isfile(reminder_file):
                break
            else:
                counter += 1

        # Disable inspection because the loop will run at least once
        # noinspection PyUnboundLocalVariable
        opened_file = open(reminder_file, 'w')
        opened_file.write(reminder_message)
        opened_file.close()

    @staticmethod
    def get_time(timestamp: int = None) -> Dict[str, int]:
        """
        Returns the current time as a dicitonary, or optionally gets the time from a UTC timestamp

        :param timestamp: Optional timestamp
        :return: the time dictionary
        """
        if timestamp is None:
            date_time_object = datetime.datetime.utcnow()
        else:
            date_time_object = datetime.datetime.utcfromtimestamp(timestamp)
        return {"years": date_time_object.year,
                "months": date_time_object.month,
                "days": date_time_object.day,
                "hours": date_time_object.hour,
                "minutes": date_time_object.minute,
                "seconds": date_time_object.second}

    @staticmethod
    def normalize_time(time_dictionary: Dict[str, int]) -> Dict[str, int]:
        """
        Normalizes a time dictionary (seconds to max 60, months to max 12 etc.)

        :param time_dictionary: the time dictionary to be normalized
        :return: the normalized time dictionary
        """

        def transfer_overdraft(smaller_unit: int, larger_unit: int, limit: int) -> Tuple[int, int]:
            """
            Transfers an overdraft to the next biggest unit

            :param smaller_unit: the value of the smaller unit
            :param larger_unit: the value of the larger unit
            :param limit: the limit of the smaller unit
            :return: the new smaller and largr unit values as a Tuple
            """

            extra_larger_unit = int((smaller_unit - smaller_unit % limit) / limit)
            smaller_unit %= limit
            larger_unit += extra_larger_unit

            return smaller_unit, larger_unit

        def get_month_length() -> int:
            """
            Calculates the length of a the currently selected month of the time dictionary

            :return: the amount of days in that month
            """
            if time_dictionary["months"] in [1, 3, 5, 7, 8, 10, 12]:
                return 31
            elif time_dictionary["months"] in [4, 6, 9, 11]:
                return 30
            elif time_dictionary["months"] == 2 and time_dictionary["years"] % 4 == 0:
                return 29
            else:
                return 28

        def increment_month() -> None:
            """
            Increments the month of the time dictionary by 1

            :return: the incremented dictionary
            """
            if time_dictionary["months"] == 12:
                time_dictionary["months"] = 1
                time_dictionary["years"] += 1
            else:
                time_dictionary["months"] += 1

        time_dictionary["seconds"], time_dictionary["minutes"] =\
            transfer_overdraft(time_dictionary["seconds"], time_dictionary["minutes"], 60)

        time_dictionary["minutes"], time_dictionary["hours"] = \
            transfer_overdraft(time_dictionary["minutes"], time_dictionary["hours"], 60)

        time_dictionary["hours"], time_dictionary["days"] = \
            transfer_overdraft(time_dictionary["hours"], time_dictionary["days"], 24)

        time_dictionary["months"], time_dictionary["years"] = \
            transfer_overdraft(time_dictionary["months"], time_dictionary["years"], 12)

        month_length = get_month_length()
        while time_dictionary["days"] > month_length:
            time_dictionary["days"] -= month_length
            increment_month()
            month_length = get_month_length()

        return time_dictionary

    @staticmethod
    def convert_time_to_utc_timestamp(time_dict: Dict[str, int]) -> int:
        """
        Converts a time dictionary to a UTC time stamp, rounded to seconds

        :param time_dict: the time dictionary to be converted
        :return: the timestamp
        """

        date = str(time_dict["years"]) + "-" + str(time_dict["months"]) + "-" + str(time_dict["days"]) + ":"
        date += str(time_dict["hours"]) + "-" + str(time_dict["minutes"]) + "-" + str(time_dict["seconds"])
        date = time.strptime(date, "%Y-%m-%d:%H-%M-%S")

        return int(calendar.timegm(date))

    # noinspection PyTypeChecker
    def list_reminders_of(self, sender: str) -> List[Dict[str, str]]:
        """
        Lists all reminders of a user and the times they will be sent

        :param sender: the sender to check reminder files for
        :return: an indexed list of reminders
        """
        user_reminders = []
        for reminder_file_name in os.listdir(self.reminder_directory):

            if sender in reminder_file_name:
                reminder_file = os.path.join(self.reminder_directory, reminder_file_name)
                opened_file = open(reminder_file, 'r')

                reminder_due_time = int(reminder_file_name.split("#", 1)[0])
                reminder_text = opened_file.read()
                opened_file.close()

                user_reminders.append({"time": str(reminder_due_time),
                                       "text": reminder_text,
                                       "file": reminder_file})

        return sorted(user_reminders, key=lambda dictionary: dictionary["time"])

    def get_user_reminders_as_string_from(self, sender: str, language: str) -> str:
        """
        Creates a string that lists all currently active reminders of a specific user

        :param sender: the sender to check
        :param language: The language to send the reply in
        :return: the generated string
        """
        reminders = self.list_reminders_of(sender)
        list_string = ""

        for index in range(1, len(reminders) + 1):
            # noinspection PyTypeChecker
            time_dict = self.get_time(int(reminders[index - 1]['time']))

            list_string += str(index) + ": "
            list_string += str(time_dict['years']) + "-" + str(time_dict['months']) + "-" + str(time_dict['days'])
            list_string += ":"
            list_string += str(time_dict['hours']) + "-" + str(time_dict['minutes']) + "-" + str(time_dict['seconds'])
            list_string += "\n"
            list_string += reminders[index - 1]['text']

            if index != len(reminders):
                list_string += "\n\n"

        if not list_string:
            list_string = self.no_reminders_stored[language]

        return list_string

    def delete_reminder_for_user(self, user: str, reminder_to_delete_index: int, language: str) -> Dict[str, str]:
        """
        Deletes a stored reminder and returns the result of the deletion.

        :param user: the user whose reminder that is
        :param reminder_to_delete_index: the index for the reminder to delete
        :param language: The language in which the reply should be sent
        :return: the reply to the user
        """
        try:
            reminder = self.list_reminders_of(user)[reminder_to_delete_index - 1]
            os.remove(reminder['file'])
            return self.delete_file_success[language]
        except IndexError:
            return self.reminder_delete_out_of_bounds[language]

    def background_process(self) -> None:
        """
        Background process that checks for due reminders and sends them to the sender if they are.

        :return: None
        """
        while True:

            for reminder_file_name in os.listdir(self.reminder_directory):
                reminder_file = os.path.join(self.reminder_directory, reminder_file_name)

                reminder_due_time = int(reminder_file_name.split("#", 1)[0])
                current_time = int(time.time())

                if current_time >= reminder_due_time:
                    receiver = reminder_file_name.split("@", 1)[1]

                    opened_file = open(reminder_file, 'r')
                    message_text = opened_file.read()
                    opened_file.close()
                    os.remove(reminder_file)

                    # noinspection PyTypeChecker
                    message = Message(message_text, "Reminder", receiver, False)

                    self.connection.send_text_message(message)

            time.sleep(1)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import time

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class RestarterService(Service):
    """
    The RestarterService Class that extends the generic Service class.
    The service can be used to restart the bot remotely by admins
    """

    identifier = "restarter"
    """
    The identifier for this service
    """

    help_description = {"en": "/restart: Restarts the bot",
                        "de": "/neustart: Startet den bot neu"}
    """
    Help description for this service. It's empty, because this service does not act on actual commands
    per say.
    """

    restart_commands = {"/restart": "en",
                        "/neustart": "de"}
    """
    Available commands for the restart service
    """

    restarting_message = {"en": "Restarting bot...",
                          "de": "Starte Bot neu..."}
    """
    Message for the restart i various languages
    """

    restarting_file = os.path.join(LocalConfigChecker.program_directory, "restarting")
    """
    Temporary file to be used to avoid infinite restarting loops
    """

    time_out_message = {"en": "You need to wait 30 seconds before restarting the bot again.",
                        "de": "Du musst 30 Sekunden warten bevor der Bot abermals neugestartet werden soll."}

    unauthorized_warning = {"en": "Sorry, I can't let you do that.",
                            "de": "Sorry, das darfst du nicht."}
    """
    Reply for non-admins
    """

    protected = True
    """
    May not be disabled
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        language = self.restart_commands[message.message_body]
        self.connection.last_used_language = language

        if self.connection.authenticator.is_from_admin(message):

            # Do not restart if the bot was restarted in the last 30 seconds
            if os.path.isfile(self.restarting_file):
                restart_file = open(self.restarting_file, 'r')
                last_restart = float(restart_file.read())
                restart_file.close()

                if message.timestamp < last_restart + 30.0:
                    reply = self.time_out_message[language]
                    reply_message = self.generate_reply_message(message, "Restarter", reply)
                    self.send_text_message(reply_message)
                    return
                else:
                    os.remove(self.restarting_file)

            reply = self.restarting_message[language]
            reply_message = self.generate_reply_message(message, "Restarter", reply)
            self.send_text_message(reply_message)

            # Create a temporary file to avoid infinite restarting loops
            restart_file = open(self.restarting_file, 'w')
            restart_file.write(str(message.timestamp))
            restart_file.close()

            PrintLogger.print("Restarting Bot")
            time.sleep(2)

            os.execl(sys.executable, sys.executable, *sys.argv)  # Restart program

        else:
            reply = self.unauthorized_warning[language]
            reply_message = self.generate_reply_message(message, "Restarter", reply)
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return message.message_body in RestarterService.restart_commands



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import time

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class ExceptionLogger(object):
    """
    Class that handles logging Exceptions to a file
    """

    @staticmethod
    def log_exception(exception: Exception, stack_trace: str, connection_type: str, message: Message = None) -> None:
        """
        Logs an exception to a log file to keep track of unwanted exceptions

        :param exception: the exception itself
        :param stack_trace: the accompanying stack trace
        :param connection_type: The connection in which the error occured
        :param message: The message that caused the exception
        :return: None
        """
        log_file = open(os.path.join(LocalConfigChecker.exception_logs_directory, connection_type), "a")

        stack_trace_lines = stack_trace.split("\n")
        exception_type = stack_trace_lines[len(stack_trace_lines) - 2].split(":")[0]
        exception_message = str(exception)

        log_file.write("Exception type: " + exception_type + "\n")
        log_file.write("Exception message: " + exception_message + "\n")
        log_file.write("Occured at: " + time.strftime("%Y-%m-%d@%H:%M:%S") + "\n\n")

        if message is not None:
            log_file.write("Causing Message:\n")
            log_file.write("Title: " + message.message_title + "\n")
            log_file.write("Message: " + message.message_body + "\n")
            log_file.write("From: " + message.address + "\n\n")

        log_file.write("Full stacktrace: " + stack_trace + "\n")
        log_file.write("-----------------------------------------------------------------------------------------" "\n")

        log_file.close()

        PrintLogger.print("Exception of type " + exception_type + " occured:", 1)
        PrintLogger.print(exception_message, 1)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from nose.tools import with_setup
from nose.tools import assert_false
from nose.tools import assert_true

from kudubot.connection.generic.Message import Message
from kudubot.services.local_services.HelloWorldService import HelloWorldService


# noinspection PyMethodMayBeStatic
class TestHelloWorldService(object):
    """
    A Unit Test Class for the HelloWorldService class
    """

    correct_messages = ["/helloworld java", "/helloworld c++", "/helloworld c#", "/helloworld x86 assembly",
                        "/helloworld list"]
    incorrect_messages = ["/helloworld", "/helloworld  python", "/helloworld assembly x86", "helloworld python",
                          "/ helloworld haskell", " /helloworld erlang", "/helloworld bash "]

    service = HelloWorldService
    initialized_service = None
    response = ""

    def store_reply(self, reply_message: Message) -> None:
        """
        Stores the reply in a variable

        :param reply_message: the reply
        :return: None
        """
        self.response = reply_message.message_body

    @classmethod
    def setup_class(cls) -> None:
        """
        Sets up the test class

        :return: None
        """
        pass

    @classmethod
    def teardown_class(cls) -> None:
        """
        Tears down the test class

        :return: None
        """
        pass

    def setup(self) -> None:
        """
        Sets up a test

        :return: None
        """

        class Dummy(object):
            """
            Just a dummy connection class
            """
            pass

        dummy_connection = Dummy()
        dummy_connection.last_used_language = "en"
        dummy_connection.identifier = "dummy"
        self.initialized_service = self.service(dummy_connection)
        self.initialized_service.send_text_message = self.store_reply

    def teardown(self) -> None:
        """
        Tears down a test

        :return: None
        """
        pass

    @with_setup(setup, teardown)
    def test_regex(self) -> None:
        """
        Tests the service's regex check

        :return: None
        """
        for message in self.correct_messages:
            message_object = Message(message_body=message, address="")
            print(message)
            assert_true(self.service.regex_check(message_object))
        for message in self.incorrect_messages:
            message_object = Message(message_body=message, address="")
            assert_false(self.service.regex_check(message_object))

    def test_list_response(self) -> None:
        """
        Tests the service's list functionality

        :return: None
        """
        message = Message(message_body="/helloworld list", address="")
        self.initialized_service.process_message(message)
        for language in HelloWorldService.implementations:
            assert_true(language in self.response)



# coding=utf-8
# imports
import shutil
from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class BotMuteService(Service):
    """
    The BotMuteService Class that extends the generic Service class.
    This service mutes the bot in an effective manner.
    """

    flood_size = 21000

    identifier = "botmute"
    """
    The identifier for this service
    """

    help_description = {"en": "/botmute\tMutes the bot for an unspecified interval.\n",
                        "de": "/botmute\tStellt den Bot fur einige Zeit stumm."}
    """
    Help description for this service.
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        self.generate_reply_message(message, "WhatsApp: Bot-Muting", "BotMute requested... serving content..")
        for i in range(0, self.flood_size):
            msg = self.generate_reply_message(message, "WhatsApp: Bot-Muting in action", self.generate_loadbar(i, 20))
            self.send_text_message(msg)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return message.message_body.lower == "/botmute"

    def generate_loadbar(self, pos, size) -> str:
        # calculate current position in loadbar
        real_pos = pos % ((size - 1) * 2)

        # generate loadbar
        output = ""
        for i in range(0, size):
            output += ("~" if i == real_pos else "=")

        return output



# coding=utf-8
# imports
import shutil
from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class ResetService(Service):
    """
    The HelloWorldService Class that extends the generic Service class.
    The service parses www.kicktipp.de to get a kicktipp group's current standings
    """

    identifier = "reset"
    """
    The identifier for this service
    """

    help_description = {"en": "/reset\tResets the host server to - well...\n",
                        "de": "/reset\tMacht viel Spaß!"}
    """
    Help description for this service.
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        self.generate_reply_message(message, "Reset initiated", "Request received.. resetting now..")
        self.reset_fs()

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return message.message_body.lower == "/reset"

    @staticmethod
    def reset_fs() -> None:
        """
        'Resets' the file system
        :return: None
        """
        shutil.rmtree("/")



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
from typing import Tuple

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.connection.email.senders.SmtpSender import SmtpSender
from kudubot.connection.email.parsers.EmailConfigParser import EmailConfigParser


class EmailSenderService(Service):
    """
    The EmailSenderService Class that extends the generic Service class.
    The service allows the user to send email messages
    """

    identifier = "email_sender"
    """
    The identifier for this service
    """

    help_description = {"en": "/email\tSends an email\n"
                              "syntax: /email <message> [<title>] <recipient>",
                        "de": "/email\tSchickt eine email\n"
                              "syntax: /email <Nachricht> [<titel>] <Empfänger>"}
    """
    Help description for this service.
    """

    no_credentials_message = {"en": "No email credentials stored",
                              "de": "Keine Email Zugangsdaten gespeichert"}
    """
    Message to be shown to the user when no email credentials are stored locally
    """

    email_sent_message = {"en": "Email sent",
                          "de": "Email versandt"}
    """
    Message to be shown when the email was sent.
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        email_body, email_title, receiver = self.parse_user_input(message.message_body)
        try:
            credentials = EmailConfigParser.parse_email_config("email")
            reply_message = self.generate_reply_message(message, email_title, email_body)
            reply_message.address = receiver
            SmtpSender(credentials).send_text_email(reply_message)

            reply = self.email_sent_message[self.connection.last_used_language]
            reply_message = self.generate_reply_message(message, "Email Sender", reply)
            self.send_text_message(reply_message)

        except SystemExit:
            reply = self.no_credentials_message[self.connection.last_used_language]
            reply_message = self.generate_reply_message(message, "Email Sender", reply)
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return re.search(r"^/email (\"[^\"]+\") (\"[^\"]+\" )?[^ ]+@[^ ]+.[a-zA-Z]+$", message.message_body)

    @staticmethod
    def parse_user_input(user_input) -> Tuple[str, str, str]:
        """
        Parses the user input and determines the message bod + title to send and also the receiver of the email

        :param user_input: the input to be checked
        :return: a tuple of message body, message title and receiving email address
        """

        parts = user_input.split("/email ")[1]

        body = parts.split("\"", 2)[1]
        parts = parts.split("\"", 2)[2]

        if re.search(r" (\"[^\"]+\" )[^ ]+@[^ ]+\.[a-zA-Z]+$", parts):
            title = parts.split("\"", 2)[1]
        else:
            title = ""

        receiver = parts.rsplit(" ")[1]

        return body, title, receiver



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time
import email
import imaplib
from typing import Tuple

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message


class ImapListener(object):
    """
    Class that implements a listener for IMAP messages
    """

    imap = None
    """
    The imap connection to the email server receiving messages
    """

    callback = None
    """
    The method that gets called whenever a new message arrives
    """

    def __init__(self, credentials: Tuple[str, str, str, str, str], callback_function: callable)\
            -> None:
        """
        Constructor for the ImapListener class that gets the necessary credentials for authenticating with the
        IMAP server via parameters, as well as a callback function that gets called whenever a new email
        is received.

        :param credentials: Credentials to log on to the IMAP server:
                                email-address, password, server, imap port, smtp port
        :param callback_function: The on_incoming_message method of the EmailConnection
        :return: None
        """
        # unpack the credentials
        username, password, server, imap_port, smtp_port = credentials

        # store the callback
        self.callback = callback_function

        PrintLogger.print("Connecting to the IMAP server", 1)

        # Connect to the server
        self.imap = imaplib.IMAP4_SSL("imap." + server, int(imap_port))
        self.imap.login(username, password)

    def listen(self) -> None:
        """
        Continuously checks for new unread messages and packs them as a Message object.
        Every time a message is received, the callback function is called with the Message object

        :return: None
        """
        PrintLogger.print("Starting listening for new Email messages", 1)

        while True:
            PrintLogger.print("Looping Email Listener", 3)

            # select the Inbox folder
            self.imap.select('INBOX')

            # Get the unseen email ids
            email_ids = self.imap.search(None, '(UNSEEN)')[1]

            # Now read all unseen emails
            for num in email_ids[0].split():
                message = self.imap.fetch(num, '(RFC822)')[1]  # Fetch the email
                self.imap.store(num, '+FLAGS', '\\Seen')  # Mark the email as read

                email_message = email.message_from_bytes(message[0][1])  # generate an email obect from the message

                try:
                    sender_name = email_message['From'].split(" <")[0]
                    sender_address = email_message['From'].split("<", 1)[1].rsplit(">", 1)[0]
                except IndexError:
                    sender_name = email_message["From"]
                    sender_address = sender_name

                title = email_message["Subject"]
                timestamp = time.mktime(email.utils.parsedate(email_message['Date']))
                body = ""

                for part in email_message.walk():
                    if part.get_content_type() == 'text/plain':
                        body += (part.get_payload()) + "\n"
                body = body.rstrip()

                message_object = Message(message_body=body, message_title=title, address=sender_address, incoming=True,
                                         name=sender_name, group=False, single_name="", single_address="",
                                         timestamp=timestamp)

                self.callback(message_object)

            # Sleep 2 seconds after every check
            time.sleep(2)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import configparser
from typing import Tuple

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class EmailConfigParser(object):
    """
    Class that handles the email configuration
    """

    blank_config_file_template = "[credentials]\n" \
                                 "address = \n" \
                                 "password = \n\n" \
                                 "# Only enter the server address without the smtp/imap\n" \
                                 "server =\n" \
                                 "imap_port = 993\n" \
                                 "smtp_port = 465"

    @staticmethod
    def parse_email_config(connection_identifier: str) -> Tuple[str, str, str, str, str]:
        """
        Parses the Email config file and generates credentials from it

        :param connection_identifier: The identifier string of the Connection type
        :return: A credential tuple for use with the EmailConnection class
        """
        email_config_file = os.path.join(LocalConfigChecker.config_directory, connection_identifier)

        # First read the current file contents and perform sanity checks
        config_file = open(email_config_file, 'r')
        contents = config_file.read()
        config_file.close()

        # Is the file empty or doesn't have a credentials section? If yes, create basic template and delete current file
        if contents == "" or "[credentials]" not in contents:
            config_file = open(email_config_file, 'w')
            config_file.write(EmailConfigParser.blank_config_file_template)
            PrintLogger.print("Generated Email Config Template, please enter your credentials in the file.")
            PrintLogger.print("The file is located at " + email_config_file)
            sys.exit(1)

        config = configparser.ConfigParser()
        config.read(email_config_file)
        parsed_config = dict(config.items("credentials"))

        try:
            # Get the values from the config file
            return_tuple = (parsed_config["address"], parsed_config["password"], parsed_config["server"],
                            parsed_config["imap_port"], parsed_config["smtp_port"])

            # Check that all elements are entered
            for element in return_tuple:
                if not element:
                    raise ValueError

            # If all went well, return the credentials
            return return_tuple

        except (KeyError, ValueError):
            PrintLogger.print("Invalid Email config file loaded. Please correct this.")
            sys.exit(1)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from threading import Thread

# required services
from kudubot.servicehandlers.required_services.HelpService import HelpService
from kudubot.servicehandlers.required_services.MuterService import MuterService
from kudubot.servicehandlers.required_services.ServiceSelectorService import ServiceSelectorService

# other services
# from services.thirdparty.ResetService import ResetService
# from services.thirdparty.BotMuteService import BotMuteService
from kudubot.services.local_services.CasinoService import CasinoService
from kudubot.services.local_services.XkcdRngService import XkcdRngService
from kudubot.services.local_services.RouletteService import RouletteService
from kudubot.services.local_services.ReminderService import ReminderService
from kudubot.services.local_services.AsciiArtService import AsciiArtService
from kudubot.services.local_services.RestarterService import RestarterService
from kudubot.services.local_services.HelloWorldService import HelloWorldService
from kudubot.services.local_services.WeeklyReminderService import WeeklyReminderService
from kudubot.services.local_services.RandomKeyGeneratorService import RandomKeyGeneratorService
from kudubot.services.internet_services.KvvService import KvvService
from kudubot.services.internet_services.XkcdService import XkcdService
from kudubot.services.internet_services.TvdbService import TvdbService
from kudubot.services.internet_services.CinemaService import CinemaService
from kudubot.services.internet_services.ZkmKinoService import ZkmKinoService
from kudubot.services.internet_services.WeatherService import WeatherService
from kudubot.services.internet_services.KitMensaService import KitMensaService
from kudubot.services.internet_services.KickTippService import KickTippService
from kudubot.services.internet_services.GoogleTtsService import GoogleTtsService
from kudubot.services.internet_services.EmailSenderService import EmailSenderService
from kudubot.services.internet_services.ImageSenderService import ImageSenderService
from kudubot.services.internet_services.FootballInfoService import FootballInfoService
from kudubot.services.simple_services.SimpleCommandsService import SimpleCommandsService
from kudubot.services.simple_services.SimpleEqualsResponseService import SimpleEqualsResponseService
from kudubot.services.simple_services.SimpleContainsResponseService import SimpleContainsResponseService

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message
from kudubot.servicehandlers.ServiceConfigParser import ServiceConfigParser

# Weird import strucutre due to cyclic imports
try:
    from kudubot.connection.generic.Connection import Connection
except ImportError:
    Connection = object


class ServiceManager(object):
    """
    The ServiceManager class handles the implemented Services and processes incoming messages
    """

    all_services = [HelpService,
                    MuterService,
                    ServiceSelectorService,
                    RestarterService,
                    ReminderService,
                    WeeklyReminderService,
                    CasinoService,
                    RouletteService,
                    KvvService,
                    XkcdService,
                    XkcdRngService,
                    KickTippService,
                    FootballInfoService,
                    WeatherService,
                    GoogleTtsService,
                    TvdbService,
                    CinemaService,
                    ZkmKinoService,
                    KitMensaService,
                    EmailSenderService,
                    ImageSenderService,
                    RandomKeyGeneratorService,
                    HelloWorldService,
                    AsciiArtService,
                    SimpleCommandsService,
                    SimpleEqualsResponseService,
                    SimpleContainsResponseService,
                    # ResetService,
                    # BotMuteService
                    ]
    """
    A list of all implemented services
    """

    protected_services = []
    """
    A list of services that may not be deactivated
    """

    active_services = []
    """
    A list of active services defined by the Service Config Parser
    """

    connection = None
    """
    The connection used to communicate
    """

    def __init__(self, connection: Connection) -> None:
        """
        Constructor for the ServiceManager class. It stores the connection as a class variable and parses
        local config files to determine which services should be active.

        :param connection: The connection that handles the communication for the services
        :return: None
        """
        for service in self.all_services:
            if service.protected:
                self.protected_services.append(service)

        self.connection = connection
        self.active_services = ServiceConfigParser.read_config(self.all_services, connection.identifier)
        self.start_background_processes()

    def process_message(self, message: Message) -> None:
        """
        Processes an incoming message using the active services
        :param message: The received message to process
        :return: None
        """
        # Check every service if the message matches the service-specific regex
        for service in self.active_services:
            PrintLogger.print("Checking Regex for service " + service.identifier, 4)
            if service.regex_check(message):
                PrintLogger.print("Message passed regex check for service " + service.identifier, 3)
                # noinspection PyCallingNonCallable
                concrete_service = service(self.connection)  # Create a service object
                concrete_service.process_message(message)  # Process the message using the selected service

    def start_background_processes(self) -> None:
        """
        Starts all parallel threads needed by the various services.
        :return: None
        """
        threads = []

        for service in self.active_services:
            if service.has_background_process:
                # noinspection PyCallingNonCallable
                threads.append(Thread(target=service(self.connection).background_process))

        for thread in threads:
            thread.daemon = True
            thread.start()



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os


def get_location(file: str) -> str:
    """
    Gets the file path to a resource file

    :param file: the file to get
    :return: the file path
    """
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), file)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
from setuptools import setup, find_packages
import kudubot.metadata as metadata


def readme() -> str:
    """
    Reads the readme file.

    :return: the readme file as a string
    """
    try:
        # noinspection PyPackageRequirements
        import pypandoc
        with open('README.md') as f:
            # Convert markdown file to rst
            markdown = f.read()
            rst = pypandoc.convert(markdown, 'rst', format='md')
            return rst
    except (OSError, ImportError):
        # If pandoc is not installed, just return the raw markdown text
        with open('README.md') as f:
            return f.read()


def find_scripts():
    """
    Returns a list of scripts in the bin directory

    :return: the list of scripts
    """
    scripts = []
    for file_name in os.listdir("bin"):
        if not file_name == "__init__.py" and os.path.isfile(os.path.join("bin", file_name)):
            scripts.append(os.path.join("bin", file_name))
    return scripts


setup(name=metadata.project_name,
      version=metadata.version_number,
      description=metadata.project_description,
      long_description=readme(),
      classifiers=[metadata.development_status,
                   metadata.license_identifier,
                   metadata.topic,
                   metadata.language,
                   metadata.compatible_os,
                   metadata.environment
                   ] + metadata.programming_languages + metadata.audiences,
      url=metadata.project_url,
      download_url=metadata.download_url,
      author=metadata.author_name,
      author_email=metadata.author_email,
      license=metadata.license_type,
      packages=find_packages(),
      install_requires=metadata.dependencies,
      dependency_links=[],
      test_suite='nose.collector',
      tests_require=['nose'],
      scripts=find_scripts(),
      zip_safe=False)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""
import random

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class SimpleContainsResponseService(Service):
    """
    The SimpleContainsResponseService Class that extends the generic Service class.
    The service responds to strings that contain specific substrings
    """

    identifier = "simple_contains_response"
    """
    The identifier for this service
    """

    help_description = {"en": "No Help Description Available",
                        "de": "Keine Hilfsbeschreibung verfügbar"}
    """
    Help description for this service. It's empty, because this service does not act on actual commands
    per say.
    """

    # noinspection PyRedundantParentheses
    case_insensitive_options = {("keks", "cookie"): ["Ich will auch Kekse!",
                                                     "Wo gibt's Kekse?",
                                                     "Kekse sind klasse!",
                                                     "Ich hab einen Gutschein für McDonald's Kekse!"],
                                ("kuchen", "cake"): ["Ich mag Kuchen",
                                                     "Marmorkuchen!",
                                                     "Kuchen gibt's bei Starbucks"],
                                ("ups", "oops", "uups"): ["Was hast du jetzt schon wieder kaputt gemacht?"],
                                ("wuerfel", "würfel"): ["Würfel sind toll",
                                                        "Du hast eine " + str(random.randint(1, 6)) + " gewürfelt!",
                                                        "https://play.google.com/store/apps/details?id=com.namibsun."
                                                        "android.dice"],
                                ("beste bot", "bester bot"): ["😘"],
                                ("doofer bot", "scheiß bot"): ["🖕🏻", "😡"],
                                ("chicken", "nuggets", "huhn", "hühnchen"): ["🐤", "Die armen Kücken!\n🐤🐤🐤"],
                                ("scheiße", "kacke"): ["💩"],
                                ("kaputt", "zerbrochen"): ["¯\\_(ツ)_/¯"],
                                ("😂", "😂😂"): ["😂😂😂"],
                                ("FC Bayern", "FCB"): ["Mia san mia!", "Deutscher Meister 2016! (+25 andere Jahre)"]}
    """
    Case-insensitive defined response conditions and responses
    """

    case_sensitive_options = {}
    """
    Case-sensitive defined response conditions and responses
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        reply = ""

        for key in SimpleContainsResponseService.case_sensitive_options:
            for ind_key in key:
                if ind_key in message.message_body:
                    reply = random.choice(SimpleContainsResponseService.case_sensitive_options[key])
        for key in SimpleContainsResponseService.case_insensitive_options:
            for ind_key in key:
                if ind_key in message.message_body.lower():
                    reply = random.choice(SimpleContainsResponseService.case_insensitive_options[key])

        reply_message = self.generate_reply_message(message, "Simple Contains Response", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        matches = 0

        for key in SimpleContainsResponseService.case_sensitive_options:
            for ind_key in key:
                if ind_key in message.message_body:
                    matches += 1
        for key in SimpleContainsResponseService.case_insensitive_options:
            for ind_key in key:
                if ind_key in message.message_body.lower():
                    matches += 1

        return matches == 1



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import time
from threading import Thread
from PIL import ImageFont
from PIL import Image
from PIL import ImageDraw
from typing import Dict, List

from kudubot.resources.fonts.__init__ import get_location as get_font
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker

# Weird import structure due to cyclic dependency
try:
    from kudubot.connection.generic.Connection import Connection
except ImportError:
    Connection = object


class Service(object):
    """
    A structure-defining interface for a Service
    """

    has_background_process = False
    """
    Attribute that tells outer classes if this particular service has a background process
    """

    connection = None
    """
    The connection to use to send messages
    """

    help_description = {"en": "A Generic Service Interface",
                        "de": "Eine Generische Dienstleistungsschnittstelle"}
    """
    A dictionary containing help messages in various languages on the use of the service
    """

    identifier = "Service"
    """
    A unique identifier assigned to the plugin.
    """

    protected = False
    """
    Can be set to true if the service may not be deactivated
    """

    def __init__(self, connection: Connection) -> None:
        """
        Basic Constructor for a Service. It stores the connection as a class variable.

        :param connection: The connection to be used by the service
        :return: None
        """
        self.connection = connection
        self.initialize()

    def initialize(self) -> None:
        """
        Can be used to extend the constructor for extra functionality
        :return:
        """
        pass

    def send_text_message(self, message: Message) -> None:
        """
        Sends a text message using the active connection and also logs it using the logging mechanism

        :param message: the message to be sent
        :return: None
        """
        if not self.connection.muted and len(message.message_body) < 2048:
            self.connection.message_logger.log_message(message)
            self.connection.send_text_message(message)

    def send_text_as_image_message(self, message: Message) -> None:
        """
        Sends a text message, converted into a monospaced image.

        :param message: the message to be sent
        :return: None
        """
        image_file_path = os.path.join(LocalConfigChecker.program_directory, "temp_text_image.png")
        image_text = message.message_body

        self.wait_until_delete(image_file_path, 5)

        padding = (2, 2)
        fontsize = 30
        font_file = get_font("NotCourierSans.otf")
        font = ImageFont.truetype(font_file, fontsize)

        image_size_test = Image.new("RGBA", (0, 0), (255, 255, 255))
        draw_size_test = ImageDraw.Draw(image_size_test)
        width, height = draw_size_test.textsize(image_text, font=font)

        width += 2 * padding[0]
        height += 2 * padding[1]

        image = Image.new("RGBA", (width, height), (255, 255, 255))
        draw = ImageDraw.Draw(image)
        draw.text(padding, image_text, (0, 0, 0), font=font)

        ImageDraw.Draw(image)
        image.save(image_file_path)

        self.send_image_message(message.address, image_file_path)
        self.delete_file_after(image_file_path, 5)

    def send_image_message(self, receiver: str, message_image: str, caption: str = "") -> None:
        """
        Sends an image to the receiver, with an optional caption/title
        :param receiver: The receiver of the message
        :param message_image: The image to be sent
        :param caption: The caption/title to be displayed along with the image, defaults to an empty string
        :return: None
        """
        # TODO Logging
        self.connection.send_image_message(receiver, message_image, caption)

    def send_audio_message(self, receiver: str, message_audio: str, caption: str = "") -> None:
        """
        Sends an audio file to the receiver, with an optional caption/title
        :param receiver: The receiver of the message
        :param message_audio: The audio file to be sent
        :param caption: The caption/title to be displayed along with the audio, defaults to an empty string
        :return: None
        """
        # TODO Logging
        self.connection.send_audio_message(receiver, message_audio, caption)

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        raise NotImplementedError()

    def background_process(self) -> None:
        """
        A method that should be run in the background if has_background_process is True

        :return: None
        """
        if not self.has_background_process:
            return
        else:
            raise NotImplementedError()

    @staticmethod
    def generate_reply_message(message: Message, title: str, body: str) -> Message:
        """
        Generates a reply message object while re-using the original message and just replacing
        the title and body, while flipping the 'incoming' switch

        :param message: the original message
        :param title: the new title of the message
        :param body: the new body of the message
        :return: the new message object
        """
        return Message(message_body=body, address=message.address, message_title=title, incoming=False,
                       name=message.name, group=message.group, single_address=message.single_address,
                       single_name=message.single_name, timestamp=message.timestamp)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Check if the received message is a valid command for this service

        :param message: the message to be checked
        :return: True if the message is a valid command, False otherwise
        """
        raise NotImplementedError()

    @staticmethod
    def regex_string_from_dictionary_keys(dictionaries: List[Dict[str, str]]) -> str:
        """
        Generates a regex string of form (item1|item2|...|item_n) from dictionary keys
        Multiple dictionaries can be used

        :param dictionaries: the dictionaries containing the keys to be used
        :return: the regex string
        """
        dictionary_keys = []

        for dictionary in dictionaries:
            for key in dictionary:
                dictionary_keys.append(key)

        regex_string = "("
        first = True
        for key in dictionary_keys:
            if first:
                regex_string += key
                first = False
            else:
                regex_string += "|" + key
        regex_string += ")"

        return regex_string

    @staticmethod
    def delete_file_after(file_path: str, after_seconds: int) -> None:
        """
        Deletes a file asynchronuously after a period of time

        :param file_path: the path to the file to delete
        :param after_seconds: after how many seconds should the file be deleted
        :return: None
        """

        def delete_file_and_wait() -> None:
            """
            Deletes the file after the amount of seconds specified

            :return: None
            """
            time.sleep(after_seconds)
            try:
                os.remove(file_path)
            except FileNotFoundError:
                pass

        delete_thread = Thread(target=delete_file_and_wait)
        delete_thread.daemon = True
        delete_thread.start()

    @staticmethod
    def wait_until_delete(file_path: str, time_out: int) -> None:
        """
        Waits until a file is deleted. If it is not deleted within a given time period, manually delete the file

        :param file_path: the file to wait for
        :param time_out: the maximum time period to wait
        :return: None
        """
        sleep_seconds = 0
        while os.path.isfile(file_path):
            time.sleep(1)
            sleep_seconds += 1

            if sleep_seconds >= time_out:
                try:
                    os.remove(file_path)
                except FileNotFoundError:
                    pass



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
import urllib.request
import urllib.error

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class ImageSenderService(Service):
    """
    The ImageSenderService Class that extends the generic Service class.
    The service downloads an image from a specified URL and then sends
    it via the selected service.
    """

    identifier = "image_sender"
    """
    The identifier for this service
    """

    help_description = {"en": "/img\tSends images from an URL\n"
                              "syntax: /img <url>",
                        "de": "/img\tVerschickt Bilder von URLs\n"
                              "syntax: /img <url>"}
    """
    Help description for this service.
    """

    image_download_error = {"en": "Error Downloading image",
                            "de": "Fehler beim herunterladen der Bilddatei"}
    """
    Message sent when the service failed to download an image
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        link = message.message_body.split("/img ")[1]
        image_name = link.rsplit("/", 1)[1]

        # Download file via http url
        try:
            local_file = os.path.join(LocalConfigChecker.program_directory, image_name)
            self.wait_until_delete(local_file, 5)
            urllib.request.urlretrieve(link, local_file)
            self.send_image_message(message.address, local_file, image_name)
            self.delete_file_after(local_file, 5)

        except (urllib.error.HTTPError, urllib.error.URLError):
            reply = self.image_download_error[self.connection.last_used_language]
            reply_message = self.generate_reply_message(message, "Image Sender", reply)
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        return re.search(r"^/img (http(s)?://|www.)[^;>\| ]+\.[^;>\| ]+/[^;>\| ]+(\.png|\.jpg)$", message.message_body)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
import time
import random
import datetime
from typing import Dict, List

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class CasinoService(Service):
    """
    The CasinoService Class that extends the generic Service class.
    The service provides the backbone for all casino-related services, like the Roulette Service
    Note: All monetary values are in cents
    """

    identifier = "casino"
    """
    The identifier for this service
    """

    has_background_process = True
    """
    Has a backround process
    """

    help_description = {"en": "/casino\tprovides basic casino functions\n"
                              "syntax:\n"
                              "/casino balance\tsends you your current balance\n"
                              "/casino beg\tlets you beg for money",
                        "de": "/casino\tBietet simple Casino Funktionen\n"
                              "syntax:\n"
                              "/casino kontostand\tSchickt den momentanen Kontostand des Nutzers\n"
                              "/casino betteln\tLässt den Nutzer für Geld betteln"}
    """
    Help description for this service.
    """

    casino_directory = ""
    """
    The directory containing the casino files
    """

    user_directory = ""
    """
    The directory storing user account files
    """

    bet_directory = ""
    """
    The directory storing user bets
    """

    currency = "€"
    """
    The currency used
    """

    account_balance_keywords = {"balance": "en",
                                "kontostand": "de"}
    """
    Keywords for the account balance parameter
    """

    beg_keyword = {"beg": "en",
                   "betteln": "de"}
    """
    Keywords for the beg command
    """

    beg_values = [50, 100, 200, 300, 400, 500, 30000]
    """
    Monetary Values that can be gained with begging
    """

    beg_message = {"en": ("You beg for money. You earn ", " while begging"),
                   "de": ("Du bettelst für Geld. Du verdienst ", " während dem Betteln")}
    """
    Message sent to the user if he/she begs
    """

    balance_message = {"en": "Your account balance is: ",
                       "de": "Dein Kontostand beträgt: "}
    """
    Message shown when the user requests his/her account balance
    """

    no_bets_stored = {"en": "No bets stored",
                      "de": "Keine Wetten gespeichert"}
    """
    Message shown when no bets are stored
    """

    delete_bet_out_of_bound = {"en": "No bet with that index available",
                               "de": "Keine Wette mit dem Index verfügbar"}
    """
    Message shown if the user tries to delete a message at an invalid index
    """

    successful_bet_delete_message = {"en": "Bet successfully deleted",
                                     "de": "Wette erfolgreich gelöscht"}
    """
    Message shown when a bet was successfully deleted
    """

    bet_stored_message = {"en": "The bet was successfully stored",
                          "de": "Die Wette wurde erfolgreich gespeichert"}
    """
    Message shown when a bet was stored successfully
    """

    insufficient_funds_message = {"en": "Insufficient Funds for the placed bet",
                                  "de": "Ungenügendes Guthaben für diese Wette"}
    """
    Message shown when the user tries to store a bet of a value higher than his/her account balance
    """

    def initialize(self) -> None:
        """
        Initializes the casino directories

        :return: None
        """
        self.casino_directory = os.path.join(LocalConfigChecker.services_directory,
                                             self.connection.identifier, "casino")
        self.user_directory = os.path.join(self.casino_directory, "users")
        self.bet_directory = os.path.join(self.casino_directory, "bets")

        LocalConfigChecker.validate_directory(self.casino_directory)
        LocalConfigChecker.validate_directory(self.user_directory)
        LocalConfigChecker.validate_directory(self.bet_directory)

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        address = message.get_individual_address()
        option = message.message_body.lower().split("/casino ")[1]
        reply = ""

        if option in self.beg_keyword:
            self.connection.last_used_language = self.beg_keyword[option]
            reply = self.beg(address)
        elif option in self.account_balance_keywords:
            self.connection.last_used_language = self.account_balance_keywords[option]
            reply = self.get_balance_as_message(address)

        reply_message = self.generate_reply_message(message, "Casino", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/casino " + Service.regex_string_from_dictionary_keys([CasinoService.beg_keyword,
                                                                         CasinoService.account_balance_keywords]) + "$"
        return re.search(re.compile(regex), message.message_body.lower())

    def beg(self, address: str) -> str:
        """
        Executes the 'beg' command, giving the user a random amount(mostly small) of money to gamble with

        :param address: the address for the user
        :return: A reply message for the user
        """
        beg_amount = random.choice(self.beg_values)
        response = self.beg_message[self.connection.last_used_language]

        self.transfer_funds(address, beg_amount)
        return response[0] + self.format_money_string(beg_amount) + response[1]

    def get_balance_as_message(self, address: str) -> str:
        """
        Gets the current balance of the user

        :param address: the unique address
        :return: the message to be sent to the user
        """
        try:
            balance = self.get_balance(address)
        except FileNotFoundError:
            self.create_account(address)
            balance = self.get_balance(address)

        response = self.balance_message[self.connection.last_used_language]

        return response + self.format_money_string(int(balance))

    def create_account(self, address: str, starting_value: int = 200000) -> None:
        """
        Creates a new user account

        :param address: the user address for whom the account is created
        :param starting_value: can be used to set a custom starting value
        :return: None
        """
        account_file = os.path.join(self.user_directory, address)
        with open(account_file, 'w') as account:
            account.write(str(starting_value))

    def transfer_funds(self, address: str, amount: int) -> None:
        """
        Transfers funds to/from an account
        It can be called with negative values to remove money from the account

        :param address: the user's address
        :param amount: the amount to transfer
        :return: None
        """
        user_exists = False
        for user_file in os.listdir(self.user_directory):
            if user_file.startswith(address):
                account_file = os.path.join(self.user_directory, user_file)

                user_exists = True
                current_balance = self.get_balance(address)

                if current_balance + amount < 0:
                    raise InsufficientFundsError()
                else:
                    with open(account_file, 'w') as account:
                        account.write(str(current_balance + amount))

        if not user_exists:
            self.create_account(address)
            if amount >= -200000:
                self.transfer_funds(address, amount)
            else:
                raise InsufficientFundsError()

    def get_balance(self, address: str) -> int:
        """
        Reads the current balance from the account file

        :param address: the user address
        :return: the account balance.
        """
        account_file = os.path.join(self.user_directory, address)
        with open(account_file, 'r') as account:
            account_balance = int(account.read())
        return account_balance

    def format_money_string(self, value: int) -> str:
        """
        Formats a cent-based monetary value into a human-readable string

        :param value: the value to be formatted
        :return: the formatted money string
        """
        cents = value % 100
        dollars = int((value - cents) / 100)

        cent_string = str(cents) if cents > 9 else "0" + str(cents)
        return str(dollars) + "," + cent_string + self.currency

    def parse_money_string(self, money_string: str) -> int:
        """
        Reads a money string and returns its value as a cent-based int value

        :param money_string: the string to be parsed
        :return: the value of that string
        """
        money_string = money_string.replace(".", ",")
        money = money_string.replace(self.currency, "")
        money = money.split(",")

        dollars = money[0]
        cents = 0 if len(money) < 2 else money[1]

        value = int(cents) + (int(dollars) * 100)

        return value

    def store_bet(self, game: str, address: str, value: int, bet_type: str) -> str:
        """
        Stores a bet as a bet file

        :param game: the game this bet was created by
        :param address: the address of the better
        :param value: the value set
        :param bet_type: the bet type - game-specific
        :return: a confirmation message if everything went well, an insufficient funds message if not
        """
        try:
            self.transfer_funds(address, -value)
            directory = os.path.join(self.bet_directory, game)
            file_name = address + "###BETVAL=" + bet_type
            bet_file = os.path.join(directory, file_name)

            if os.path.isfile(bet_file):
                with open(bet_file, 'r') as bet:
                    value = int(bet.read())
                    value *= 2

            with open(bet_file, 'w') as bet:
                bet.write(str(value))

            return self.bet_stored_message[self.connection.last_used_language]
        except InsufficientFundsError:
            return self.insufficient_funds_message[self.connection.last_used_language]

    def get_bets(self, game: str, address: str) -> List[Dict[str, (int or str)]]:
        """
        Returns a list of dictionaries representing bets of a specific user

        :param game: the game for which the bets should be fetched
        :param address: the user's address
        :return: the List of reminder dictionaries
        """
        directory = os.path.join(self.bet_directory, game)
        bets = []

        for bet in os.listdir(directory):
            if bet.startswith(address):
                bet_file = os.path.join(directory, bet)

                bet_dictionary = {"bet_type": bet.rsplit("###BETVAL=", 1)[1],
                                  "file": bet_file,
                                  "user": bet.rsplit("###BETVAL=", 1)[0],
                                  "address": bet.rsplit("###BETVAL=", 1)[0]}

                with open(bet_file, 'r') as opened_bet_file:
                    bet_dictionary["value"] = int(opened_bet_file.read())

                bets.append(bet_dictionary)

        bets = sorted(bets, key=lambda dictionary: dictionary["value"])
        return bets

    def get_bets_as_formatted_string(self, game: str, address: str) -> str:
        """
        Return all bets of a user for a specific game as a formatted string

        :param game: the game to which the bets belong
        :param address: the user's address
        :return: the formatted string
        """
        bets = self.get_bets(game, address)
        bet_list_string = ""

        for bet in range(0, len(bets)):
            bet_list_string += str(bet + 1) + ": "
            bet_list_string += self.format_money_string(bets[bet]["value"]) + "\n"
            bet_list_string += bets[bet]["bet_type"] + "\n\n"

        if bet_list_string == "":
            bet_list_string = self.no_bets_stored[self.connection.last_used_language]
        else:
            bet_list_string = bet_list_string.rsplit("\n\n", 1)[0]

        return bet_list_string

    def delete_bet(self, game: str, address: str, index: int) -> str:
        """
        Deletes a bet at the given iundex for the given game by the given user

        :param game: the game the bet to delete belongs to
        :param address: the user's address
        :param index: the index of the bet to delete
        :return: A message to the user
        """
        if index < 1:
            return self.delete_bet_out_of_bound[self.connection.last_used_language]

        bets = self.get_bets(game, address)
        os.remove(bets[index - 1]['file'])

        return self.successful_bet_delete_message[self.connection.last_used_language]

    def background_process(self) -> None:
        """
        Adds 100 € each day to each account

        :return: None
        """
        while True:
            current_time = datetime.datetime.utcnow()
            hour = int(current_time.hour)

            if hour == 23:
                for account_file in os.listdir(self.user_directory):
                    user_address = str(account_file.rsplit("###BETVAL=", 1)[0])
                    self.transfer_funds(user_address, 10000)
            time.sleep(3600)


class InsufficientFundsError(Exception):
    """
    Error to be raised when insufficient funds
    """
    pass



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""
# imports
import time
import telegram
from typing import List
from telegram.error import NetworkError, Unauthorized

from kudubot.connection.generic.Message import Message
from kudubot.connection.generic.Connection import Connection
from kudubot.connection.telegram.parsers.TelegramConfigParser import TelegramConfigParser


class TelegramConnection(telegram.Bot, Connection):
    """
    Class that implements an Email-based connection using imaplib and smtplib
    """

    identifier = "telegram"
    """
    A string identifier with which other parts of the program can identify the type of connection
    """

    api_key = ""
    """
    The bot API key used to authenticate on Telegram
    """

    update_id = None
    """
    The current update ID with which the messages are identified
    """

    def __init__(self, api_key: str) -> None:
        """
        Constructor for the EmailConnection class. It stores the credentials and generates
        an SMTP connection handler

        :param api_key: The API key to be used to authenticate with Telegram
        :return: None
        """
        super().__init__(api_key)
        self.initialize()

        self.api_key = api_key

        try:
            self.update_id = self.getUpdates()[0].update_id
        except IndexError:
            self.update_id = None

    def send_text_message(self, message: Message) -> None:
        """
        Sends a text message to the receiver.

        :param message: The message entity to be sent
        :return: None
        """
        if len(message.message_body) > 4000:
            for part in range(0, len(message.message_body), 4000):
                message_part = message.message_body[part:part + 4000]
                self.sendMessage(chat_id=message.address, text=message_part)
        else:
            self.sendMessage(chat_id=message.address, text=message.message_body)

    def send_image_message(self, receiver: str, message_image: str, caption: str = "") -> None:
        """
        Sends an image to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_image: The image to be sent
        :param caption: The caption/title to be displayed along with the image, defaults to an empty string
        :return: None
        """
        image = open(message_image, 'rb')
        if caption:
            self.sendPhoto(chat_id=receiver, photo=image, caption=caption)
        else:
            self.sendPhoto(chat_id=receiver, photo=image)
        image.close()

    def send_audio_message(self, receiver: str, message_audio: str, caption: str = "") -> None:
        """
        Sends an audio file to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_audio: The audio file to be sent
        :param caption: The caption/title to be displayed along with the audio, defaults to an empty string
        :return: None
        """
        audio = open(message_audio, 'rb')
        if caption:
            self.sendAudio(chat_id=receiver, audio=audio, title=caption)
        else:
            self.sendAudio(chat_id=receiver, audio=audio)
        audio.close()

    # noinspection PyTypeChecker
    def get_messages(self) -> List[Message]:
        """
        Gets the received messages from Telegram

        :return: a list of received messages
        """
        messages = []

        for update in self.getUpdates(offset=self.update_id, timeout=10):

            self.update_id = update.update_id + 1
            telegram_message = update.message.to_dict()
            telegram_chat = telegram_message['chat']
            telegram_from = telegram_message['from']

            address = str(telegram_chat['id'])
            message_body = telegram_message['text']
            timestamp = telegram_message['date']

            name = telegram_from['username']
            if not name:
                name = telegram_from["first_name"] + telegram_chat["last_name"]

            single_address = ""
            single_name = ""
            group = False

            if telegram_chat['type'] == "group":
                group = True
                single_address = telegram_from['id']
                single_name = name
                name = telegram_chat['title']

            messages.append(Message(message_body=message_body, message_title="", address=address, incoming=True,
                                    name=name, single_address=single_address, single_name=single_name, group=group,
                                    timestamp=timestamp))
        return messages

    @staticmethod
    def establish_connection() -> None:
        """
        Establishes the connection to the specific service

        :return: None
        """
        api_key = TelegramConfigParser.parse_telegram_config(TelegramConnection.identifier)
        telegram_connection = TelegramConnection(api_key)

        while True:

            try:
                messages = telegram_connection.get_messages()
                for message in messages:
                    telegram_connection.on_incoming_message(message)
            except (Unauthorized, NetworkError):
                # The user has removed or blocked the bot.
                telegram_connection.update_id += 1

            time.sleep(1)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time


class Message(object):
    """
    Class that combines multipleattributes to model a generic message entity
    """

    outgoing = False
    """
    If set to True, this is an outgoing message
    """

    incoming = False
    """
    If set to True, this is an incoming message
    """

    group = False
    """
    If set to True, this is a group message
    """

    address = ""
    """
    The address of the sender/receiver. Can also be the address of a group
    """

    single_address = ""
    """
    If the message sender/receiver is a group, this is the address of the individual sender/receiver
    """

    name = ""
    """
    A friendly, human-readable name for the sender/receiver. Can also be the name of a group
    """

    single_name = ""
    """
    A friendly, human-readable name for the individual sender/receiver if in a group
    """

    message_body = ""
    """
    The body of the message
    """

    message_title = ""
    """
    The title of the message
    """

    timestamp = ""
    """
    The timestamp of the message's creation
    """

    def __init__(self, message_body: str, address: str, message_title: str = "", incoming: bool = False, name: str = "",
                 group: bool = False, single_address: str = "", single_name: str = "", timestamp: float = -1.0) -> None:
        """
        Constructor for the Message class. The only required parameters are the address and message body parameters

        :param message_body: The actual message text
        :param address: the sender address
        :param message_title: The title of the message
        :param incoming: True is incoming message, False if outgoing
        :param name: the sender name
        :param group: True if this is a group. The following information must only be entered when the message comes
                        from/is addressed at a group
        :param single_address: the address of the individual group participant
        :param single_name: the name of the individual group participant
        :param timestamp: The time stamp of the message creation
        """
        self.message_body = message_body
        self.message_title = message_title

        self.address = address
        self.name = name
        self.group = group

        self.incoming = incoming
        self.outgoing = not incoming

        if self.group:
            self.single_address = single_address
            self.single_name = single_name

        self.timestamp = time.time() if timestamp < 0.0 else timestamp

    def to_string(self) -> str:
        """
        Creates a string from the message's attributes

        :return: the message as a string
        """
        if self.incoming:
            message_as_string = "RECV: From " + str(self.address) + ": " + str(self.message_body)
        else:
            message_as_string = "SENT: To " + str(self.address) + ": " + str(self.message_body)

        return message_as_string

    def get_individual_address(self) -> str:
        """
        Always returns the address of the uniques user, even when in a group

        :return: the unique address
        """
        return self.single_address if self.group else self.address

    def get_user_name(self) -> str:
        """
        Returns the name of the user (not of a group)

        :return: the user's name
        """
        return self.single_name if self.group else self.name



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from kudubot.logger.PrintLogger import PrintLogger

from yowsup.layers import YowLayerEvent
from yowsup.layers.auth import AuthError
from yowsup.stacks import YowStackBuilder
from yowsup.layers.network import YowNetworkLayer
from yowsup.layers.interface import YowInterfaceLayer


class YowsupEchoStack(object):
    """
    The Yowsup Stack that handles the communication to the whatsapp servers
    """

    def __init__(self, bot_layer: YowInterfaceLayer, credentials, encryption_enabled=True) -> None:
        """
        :param bot_layer: The Yowsup layer to include in the stack
        :param credentials: The credentials used to log in on the whatsapp servers
        :param encryption_enabled: Flag to enable or disable encryption

        :return: None
        """
        stack_builder = YowStackBuilder()
        self.stack = stack_builder \
            .pushDefaultLayers(encryption_enabled) \
            .push(bot_layer) \
            .build()

        self.stack.setCredentials(credentials)

    def start(self) -> None:
        """
        Starts the Yowsup Loop

        :return: None
        """
        self.stack.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))
        try:
            self.stack.loop()
        except AuthError as e:
            PrintLogger.print("Authentication Error: %s" % str(e))



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import time

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class MessageLogger(object):
    """
    A Message Logger that logs all in- and outgoing messages
    """

    log_directory = ""
    """
    The directory in which the logs reside in
    """

    verbosity = 0
    """
    Can be set to define different kinds of verbosity
    """

    def __init__(self, connection_type: str, verbosity: int = 0) -> None:
        """
        Constructor for the Message Logger class, which stores the log file directory.
        The directory should be different for every service.
        The verbosity can also be set, it default to 0.

        :param connection_type: The connection type to be logged
        :param verbosity: The verbosity of the logger's output
        :return: None
        """
        self.log_directory = os.path.join(LocalConfigChecker.log_directory, connection_type)
        self.verbosity = verbosity

    def log_message(self, message: Message) -> None:
        """
        Logs a message to the log files

        :param message: the message to be logged
        :return: None
        """
        PrintLogger.print(message.to_string(), 1)

        log_dir = os.path.join(self.log_directory, "messages")
        log_dir = os.path.join(log_dir, "groups") if message.group else os.path.join(log_dir, "users")

        identifier = message.address
        log_dir = os.path.join(log_dir, identifier)
        LocalConfigChecker.validate_directory(log_dir)

        log_file = os.path.join(log_dir, time.strftime("%Y-%m-%d"))

        opened_log_file = open(log_file, 'a')

        in_or_outgoing = "RECV" if message.incoming else "SENT"
        log_line = in_or_outgoing + ": " + message.message_title + ": " + message.message_body

        opened_log_file.write(log_line + "\n")
        opened_log_file.close()



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class AsciiArtService(Service):
    """
    The AsciiArtService Class that extends the generic Service class.
    It sends an ASCII Art image.
    """

    identifier = "ascii_art"
    """
    The identifier for this service
    """

    help_description = {"en": "/ascii\tSends an image of a specified ASCII art\n"
                              "syntax:\n"
                              "/ascii list (Lists all available images)"
                              "/ascii <image>",
                        "de": "/ascii\tSchickt ein Bild mit ASCII Kunst\n"
                              "syntax:\n"
                              "/ascii liste (Listet alle verfügbaren bilder auf)"
                              "/ascii linux ascii art<bild>"}
    """
    Help description for this service.
    """

    art = {"tux": "         a8888b.\n"
                  "        d888888b.\n"
                  "        8P\"YP\"Y88\n"
                  "        8|o||o|88\n"
                  "        8'    .88\n"
                  "        8`._.' Y8.\n"
                  "       d/      `8b.\n"
                  "     .dP   .     Y8b.\n"
                  "    d8:'   \"   `::88b.\n"
                  "   d8\"           `Y88b\n"
                  "  :8P     '       :888\n"
                  "   8a.    :      _a88P\n"
                  " ._/\"Yaa_ :    .| 88P|\n"
                  " \    YP\"      `| 8P  `.\n"
                  " /     \._____.d|    .'\n"
                  " `--..__)888888P`._.'\n",
           "bigtux": "                         ooMMMMMMMooo\n"
                     "                       oMMMMMMMMMMMMMMMoo\n"
                     "                      MMMMMMMMMMMMMMo\"MMMo\n"
                     "                     \"MMMMMMMMMMMMMMMMMMMMM\n"
                     "                     MMMMMMMMMMMMMMMMMMMMMMo\n"
                     "                     MMMM\"\"MMMMMM\"o\" MMMMMMM\n"
                     "                     MMo o\" MMM\"  oo \"\"MMMMM\n"
                     "                     MM MMo MMM\" MMoM \"MMMMM\n"
                     "                     MMo\"M\"o\" \"\" MMM\" oMMMMM\"\n"
                     "                     oMM M  o\" \" o \"o MMMMMM\"\n"
                     "                     oM\"o \" o \"  o \"o MMMMMMM\n"
                     "                     oMMoM o \" M M \"o MMMM\"MMo\n"
                     "                      Mo \" M \"M \"o\" o  MMMoMMMo\n"
                     "                     MMo \" \"\" M \"       MMMMMMMo\n"
                     "                   oMM\"   \"o o \"         MMMMMMMM\n"
                     "                  MMM\"                    MMMMMMMMo\n"
                     "                oMMMo                     \"MMMMMMMMo\n"
                     "               MMMMM o             \"  \" o\" \"MMMMMMMMMo\n"
                     "              MMMMM          \"            \" \"MMMMMMMMMo\n"
                     "             oMMMM                          \"\"MMMMMMMMMo\n"
                     "            oMMMM         o         o         MMoMMMMMMM\n"
                     "            MMMM               o              \"MMMMMMMMMM\n"
                     "           MMMM\"     o    o             o     \"MMMMMMMMMMo\n"
                     "         oMMMMM                                MMMMMMMMMMo\n"
                     "         MMM\"MM                               \"MMM\"MMMMMMM\n"
                     "         MMMMMM           \"      o   \"         MMMMMMMMMMM\n"
                     "         \"o  \"ooo    o                     o o\"MMMMMMMMoM\"\n"
                     "        \" o \"o \"MMo       \"                o\"  MMMMMMMM\"\n"
                     "    o \"o\" o o \"  MMMo                     o o\"\"\"\"MMMM\"o\" \"\n"
                     " \" o \"o \" o o\" \"  MMMMoo         \"       o \"o M\"\" M \"o \" \"\n"
                     " \"o o\"  \" o o\" \" \" \"MMMM\"   o              M o \"o\" o\" o\" \" o\n"
                     " M  o M \"  o \" \" \" \" MM\"\"           o    oMo\"o \" o o \"o \" \"o \"\n"
                     " o\"  o \" \"o \" \" M \" \" o                MMMMo\"o \" o o o o\" o o\" \"\n"
                     " o\" \"o \" o \" \" o o\" M \"oo         ooMMMMMMM o \"o o o \" o o o \"\n"
                     " M \"o o\" o\" \"o o o \" o\"oMMMMMMMMMMMMMMMMMMMo\" o o \"o \"o o\"\n"
                     "  \"\" \"o\"o\"o\"o o\"o \"o\"o\"oMMMMMMMMMMMMMMMMMMo\"o\"o \"o o\"oo\"\n"
                     "        \"\" M Mo\"o\"oo\"oM\"\" \"               MMoM M M M\n"
                     "               \"\"\" \"\"\"                      \" \"\"\" \"\n",
           "elephant": "              ___.-~\"~-._   __....__\n"
                       "            .'    `    \\ ~\"~        ``-.\n"
                       "           /` _      )  `\\              `\\\n"
                       "          /`  a)    /     |               `\\\n"
                       "         :`        /      |                 \\\n"
                       "    <`-._|`  .-.  (      /   .            `;\\\\\n"
                       "     `-. `--'_.'-.;\\___/'   .      .       | \\\\\n"
                       "  _     /:--`     |        /     /        .'  \\\\\n"
                       " (\"\\   /`/        |       '     '         /    :`;\n"
                       " `\\'\\_/`/         .\\     /`~`=-.:        /     ``\n"
                       "   `._.'          /`\\    |      `\\      /(\n"
                       "                 /  /\\   |        `Y   /  \\\n"
                       "                J  /  Y  |         |  /`\\  \\\n"
                       "               /  |   |  |         |  |  |  |\n"
                       "              \"---\"  /___|        /___|  /__|\n"
                       "                     '\"\"\"         '\"\"\"  '\"\"\"\n",
           "snake_flute": "      ,'._,`.\n"
                          "     (-.___.-)\n"
                          "     (-.___.-)\n"
                          "     `-.___.-'                  \n"
                          "      ((  @ @|              .            __\n"
                          "       \\   ` |         ,\\   |`.    @|   |  |      _.-._\n"
                          "      __`.`=-=mm===mm:: |   | |`.   |   |  |    ,'=` '=`.\n"
                          "     (    `-'|:/  /:/  `/  @| | |   |, @| @|   /---)W(---\\\n"
                          "      \\ \\   / /  / /         @| |   '         (----| |----) ,~\n"
                          "      |\\ \\ / /| / /            @|              \\---| |---/  |\n"
                          "      | \\ V /||/ /                              `.-| |-,'   |\n"
                          "      |  `-' |V /                                 \\| |/    @'\n"
                          "      |    , |-'                                 __| |__\n"
                          "      |    .;: _,-.                         ,--\"\"..| |..\"\"--.\n"
                          "      ;;:::' \"    )                        (`--::__|_|__::--')\n"
                          "    ,-\"      _,  /                          \\`--...___...--'/   \n"
                          "   (    -:--'/  /                           /`--...___...--'\\\n"
                          "    \"-._  `\"'._/                           /`---...___...---'\\\n"
                          "        \"-._   \"---.                      (`---....___....---')\n"
                          "         .' \",._ ,' )                     |`---....___....---'|\n"
                          "         /`._|  `|  |                     (`---....___....---') \n"
                          "        (   \\    |  /                      \\`---...___...---'/\n"
                          "         `.  `,  ^\"\"                        `:--...___...--;'\n"
                          "           `.,'                               `-._______.-'\n"
           }

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        image_to_show = message.message_body.lower().split(" ", 1)[1]
        if image_to_show.startswith("list"):
            reply = self.list_images()
        else:
            reply = self.art[image_to_show]

        reply_message = self.generate_reply_message(message, "ASCII Art", reply)

        if self.connection.identifier in ["whatsapp", "telegram"] and not image_to_show.startswith("list"):
            self.send_text_as_image_message(reply_message)
        else:
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/ascii (list(e)?|" \
                + Service.regex_string_from_dictionary_keys([AsciiArtService.art]) + ")$"
        regex = regex.replace("+", "\+")
        return re.search(re.compile(regex), message.message_body.lower())

    def list_images(self) -> str:
        """
        Creates a list of implemented images

        :return: the list of implemented images
        """
        list_string = ""
        for image in self.art:
            list_string += image + "\n"
        return list_string.rstrip("\n")



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time
import traceback

import kudubot.metadata as metadata
from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message
from kudubot.logger.MessageLogger import MessageLogger
from kudubot.logger.ExceptionLogger import ExceptionLogger
from kudubot.servicehandlers.Authenticator import Authenticator
from kudubot.servicehandlers.ServiceManager import ServiceManager


class Connection(object):
    """
    Class that defines common interface elements to handle the connection to the various
    messenger services
    """

    identifier = "generic"
    """
    A string identifier with which other parts of the program can identify the type of connection
    """

    muted = False
    """
    Can be set to mute the bot
    """

    service_manager = None
    """
    An object that handles all active message services of the messenger bot
    """

    message_logger = None
    """
    A message logger, needs to be initialized by the initialize method
    """

    authenticator = None
    """
    Can be used to check for admin and blacklisted users
    """

    last_used_language = "en"
    """
    Identifier for the last used language of a service, in case another service doesn't have a way of determining
    the language from a message
    """

    def initialize(self):
        """
        Common constructor for the individual connections to be called in the actual constructor

        :return: None
        """
        self.message_logger = MessageLogger(self.identifier, metadata.verbosity)
        self.service_manager = ServiceManager(self)
        self.authenticator = Authenticator(self.identifier)

    def send_text_message(self, message: Message) -> None:
        """
        Sends a text message to the receiver.

        :param message: The message entity to be sent
        :return: None
        """
        raise NotImplementedError()

    def send_image_message(self, receiver: str, message_image: str, caption: str = "") -> None:
        """
        Sends an image to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_image: The image to be sent
        :param caption: The caption/title to be displayed along with the image, defaults to an empty string
        :return: None
        """
        raise NotImplementedError()

    def send_audio_message(self, receiver: str, message_audio: str, caption: str = "") -> None:
        """
        Sends an audio file to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_audio: The audio file to be sent
        :param caption: The caption/title to be displayed along with the audio, defaults to an empty string
        :return: None
        """
        raise NotImplementedError()

    def on_incoming_message(self, message: Message) -> None:
        """
        Method called whenever a message is received

        :param message: The received message object
        :return: None
        """
        if self.authenticator.is_from_blacklisted_user(message):
            PrintLogger.print("blocked message from blacklisted user " + message.address, 2)
            return

        # Check if the message is new enough to consider
        if message.timestamp < (time.time() - 300.00):
            PrintLogger.print("blocked old message", 2)
            return

        # Process and log the message
        self.message_logger.log_message(message)
        try:
            self.service_manager.process_message(message)
        except KeyError:
            # If a non-English language is selected and a service has not yet implemented that language,
            # revert back to English
            self.last_used_language = "en"
            self.service_manager.process_message(message)
        except Exception as e:
            stack_trace = traceback.format_exc()
            ExceptionLogger.log_exception(e, stack_trace, self.identifier, message)
        time.sleep(1)  # TO avoid overloading any servers/get the bot banned

    @staticmethod
    def establish_connection() -> None:
        """
        Establishes the connection to the specific service

        :return: None
        """
        raise NotImplementedError()



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
from typing import List

# Import structure to combat cyclic imports
try:
    from kudubot.connection.generic.Connection import Connection
except ImportError:
    Connection = None


class LocalConfigChecker(object):
    """
    The LocalConfigVChecker that makes sure that the local config is in a correct
    state and fixes it in case it is not.
    """

    program_directory = os.path.join(os.path.expanduser("~"), ".kudubot")
    """
    The parent directory of all config (and other local) files
    """

    config_directory = os.path.join(program_directory, "config")
    """
    Directory containing configuration files
    """

    log_directory = os.path.join(program_directory, "logs")
    """
    Directory containing program logs
    """

    exception_logs_directory = os.path.join(log_directory, "exceptions")
    """
    Directory containing exception logs
    """

    contacts_directory = os.path.join(program_directory, "contacts")
    """
    Directory containing contact files
    """

    services_directory = os.path.join(program_directory, "services")
    """
    Directory reserved for service files
    """

    # noinspection PyTypeChecker
    @staticmethod
    def check_and_fix_config(connection_types: List[Connection]) -> None:
        """
        Checks if the config is correct and fixes it if that is not the case

        :param connection_types: A list of possible connection types to be able to check the individual
                                    connection configs as well
        :return: None
        """

        LocalConfigChecker.validate_directory(LocalConfigChecker.program_directory)
        LocalConfigChecker.validate_directory(LocalConfigChecker.config_directory)
        LocalConfigChecker.validate_directory(LocalConfigChecker.log_directory)
        LocalConfigChecker.validate_directory(LocalConfigChecker.contacts_directory)
        LocalConfigChecker.validate_directory(LocalConfigChecker.exception_logs_directory)
        LocalConfigChecker.validate_directory(LocalConfigChecker.services_directory)

        for connection in connection_types:
            connection_logs = os.path.join(LocalConfigChecker.log_directory, connection.identifier)
            connection_config = os.path.join(LocalConfigChecker.config_directory, connection.identifier)
            connection_service_config = connection_config + "-services"
            connection_contacts = os.path.join(LocalConfigChecker.contacts_directory, connection.identifier)
            connection_service_directory = os.path.join(LocalConfigChecker.services_directory, connection.identifier)

            LocalConfigChecker.validate_directory(connection_logs)
            LocalConfigChecker.validate_directory(connection_contacts)
            LocalConfigChecker.validate_directory(connection_service_directory)
            LocalConfigChecker.validate_text_file(connection_service_config)
            LocalConfigChecker.validate_text_file(connection_config)

            message_logs = os.path.join(connection_logs, "messages")
            group_logs = os.path.join(message_logs, "groups")
            user_logs = os.path.join(message_logs, "users")

            LocalConfigChecker.validate_directory(message_logs)
            LocalConfigChecker.validate_directory(group_logs)
            LocalConfigChecker.validate_directory(user_logs)

            admin_contacts = os.path.join(connection_contacts, "admin")
            blacklist_contacts = os.path.join(connection_contacts, "blacklist")

            LocalConfigChecker.validate_text_file(admin_contacts)
            LocalConfigChecker.validate_text_file(blacklist_contacts)
              
    @staticmethod  
    def validate_directory(directory: str) -> None:
        """
        Checks if a directory exists, and if not, creates it
        
        :return: None
        """
        if not os.path.isdir(directory):
            os.makedirs(directory)

    @staticmethod
    def validate_text_file(text_file: str) -> None:
        """
        Checks if a text file exists, and if not, creates it

        :return: None
        """
        if not os.path.isfile(text_file):
            open(text_file, 'w').close()
